;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "xcscope" "20260823.51"
  "Cscope interface for (X)Emacs."
  ()
  :url "https://github.com/dkogan/xcscope.el"
  :commit "e27bf918978ac9d5e3383997dcb49d88d6135d91"
  :revdesc "e27bf918978a"
  :keywords '("languages" "c")
  :authors '(("Darryl Okahata" . "darrylo@sonic.net")
             ("Dima Kogan" . "dima@secretsauce.net"))
  :maintainers '(("Dima Kogan" . "dima@secretsauce.net")))
